from ..datastore.core import STORAGE_FORMATS

DEFAULT_NAME = "mfchckpt"
MAX_HASH_LEN = 12
CHECKPOINTS_STORAGE_PREFIX = "mf.checkpoints"
CHECKPOINT_TASK_IDENTIFIER_ENV_VAR_NAME = "METAFLOW_CHECKPOINT_TASK_IDENTIFIER"
CHECKPOINT_UID_ENV_VAR_NAME = "METAFLOW_CHECKPOINT_UID"

TASK_CHECKPOINTS_ARTIFACT_NAME = "_task_checkpoints"
TASK_LATEST_CHECKPOINT_ARTIFACT_NAME = "_task_latest_checkpoint"

CHECKPOINT_TAG_PREFIX = "checkpoint:"

DEFAULT_STORAGE_FORMAT = STORAGE_FORMATS.FILES
